@@ -711,21 +711,6 @@ struct uvm_gpu_struct
             int node_id;
         } numa;
 
-        // Coherent Driver-based Memory Management (CDMM) is a mode that allows
-        // coherent GPU memory to be managed by the driver and not the OS. This
-        // is done by the driver not onlining the memory as NUMA nodes. Having
-        // the field provides the most flexibility and is sync with the numa
-        // properties above. CDMM as a property applies to the entire system.
-        bool cdmm_enabled;
-
-        // Physical address of the start of statically mapped fb memory in BAR1
-        NvU64 static_bar1_start;
-
-        // Size of statically mapped fb memory in BAR1.
-        NvU64 static_bar1_size;
-
-        // Whether or not RM has iomapped the region write combined.
-        NvBool static_bar1_write_combined;
     } mem_info;
 
     struct
@@ -941,9 +926,6 @@ struct uvm_gpu_struct
     // Force pushbuffer's GPU VA to be >= 1TB; used only for testing purposes.
     bool uvm_test_force_upper_pushbuffer_segment;
 
-    // Have we initialised device p2p pages.
-    bool device_p2p_initialised;
-
     // Used to protect allocation of p2p_mem and assignment of the page
     // zone_device_data fields.
     uvm_mutex_t device_p2p_lock;
@@ -1014,10 +996,28 @@ struct uvm_parent_gpu_struct
     // Total amount of physical memory available on the parent GPU.
     NvU64 max_allocatable_address;
 
-#if UVM_IS_CONFIG_HMM()
+#if UVM_IS_CONFIG_HMM() || defined(NV_MEMORY_DEVICE_COHERENT_PRESENT)
     uvm_pmm_gpu_devmem_t *devmem;
 #endif
 
+    // Physical address of the start of statically mapped fb memory in BAR1
+    NvU64 static_bar1_start;
+
+    // Size of statically mapped fb memory in BAR1.
+    NvU64 static_bar1_size;
+
+    // Whether or not RM has iomapped the region write combined.
+    NvBool static_bar1_write_combined;
+
+    // Have we initialised device p2p pages.
+    bool device_p2p_initialised;
+
+    // Coherent Driver-based Memory Management (CDMM) is a mode that allows
+    // coherent GPU memory to be managed by the driver and not the OS. This
+    // is done by the driver not onlining the memory as NUMA nodes. CDMM as a
+    // property applies to the entire system.
+    bool cdmm_enabled;
+
     // The physical address range addressable by the GPU
     //
     // The GPU has its NV_PFB_XV_UPPER_ADDR register set by RM to
@@ -1365,6 +1365,20 @@ struct uvm_parent_gpu_struct
         NvU64 base_address;
     } egm;
 
+    // Peer VIDMEM base offset used when creating GPA PTEs for
+    // peer mappings. RM will set this offset on systems where
+    // peer accesses are not zero-based (NVLINK 5+).
+    struct
+    {
+        // Is the GPU directly connected to peer GPUs.
+        bool is_direct_connected;
+
+        // 48-bit fabric memory physical offset that peer gpus need in order
+        // access to be rounted to the correct peer.
+        // Each memory window is 4TB, so the upper 6 bits are used for rounting.
+        NvU64 peer_gpa_memory_window_start;
+    } peer_address_info;
+
     uvm_test_parent_gpu_inject_error_t test;
 
     // PASID ATS
@@ -1619,6 +1633,8 @@ uvm_aperture_t uvm_gpu_egm_peer_aperture(uvm_parent_gpu_t *local_gpu, uvm_parent
 
 bool uvm_parent_gpus_are_nvswitch_connected(const uvm_parent_gpu_t *parent_gpu0, const uvm_parent_gpu_t *parent_gpu1);
 
+bool uvm_parent_gpus_are_direct_connected(const uvm_parent_gpu_t *parent_gpu0, const uvm_parent_gpu_t *parent_gpu1);
+
 static bool uvm_gpus_are_smc_peers(const uvm_gpu_t *gpu0, const uvm_gpu_t *gpu1)
 {
     UVM_ASSERT(gpu0 != gpu1);
@@ -1867,6 +1883,6 @@ typedef enum
 } uvm_gpu_buffer_flush_mode_t;
 
 // PCIe BAR containing static framebuffer memory mappings for PCIe P2P
-int uvm_device_p2p_static_bar(uvm_gpu_t *gpu);
+int uvm_device_p2p_static_bar(uvm_parent_gpu_t *gpu);
 
 #endif // __UVM_GPU_H__
